@@ -6,7 +6,7 @@
 ;; URL: http://github.com/s-kostyaev/ellama
 ;; Keywords: help local tools
 ;; Package-Requires: ((emacs "28.1") (llm "0.6.0") (spinner "1.7.4"))
-;; Version: 0.3.2
+;; Version: 0.4.13
 ;; SPDX-License-Identifier: GPL-3.0-or-later
 ;; Created: 8th Oct 2023
 
@@ -40,31 +40,35 @@
 (require 'spinner)
 (eval-when-compile (require 'rx))
 
+(defgroup ellama nil
+  "Tool for interacting with LLMs."
+  :group 'tools)
+
 (defcustom ellama-buffer "*ellama*"
   "Default ellama buffer."
-  :group 'tools
+  :group 'ellama
   :type 'string)
 
 (defcustom ellama-user-nick "User"
   "User nick in logs."
-  :group 'tools
+  :group 'ellama
   :type 'string)
 
 (defcustom ellama-assistant-nick "Ellama"
   "Assistant nick in logs."
-  :group 'tools
+  :group 'ellama
   :type 'string)
 
 (defcustom ellama-buffer-mode (if (fboundp 'markdown-mode)
 				  'markdown-mode
 				'text-mode)
   "Major mode for ellama logs."
-  :group 'tools
+  :group 'ellama
   :type 'function)
 
 (defcustom ellama-language "English"
   "Language for ellama translation."
-  :group 'tools
+  :group 'ellama
   :type 'string)
 
 (defcustom ellama-provider
@@ -74,21 +78,91 @@
     (make-llm-ollama
      :chat-model "zephyr" :embedding-model "zephyr"))
   "Backend LLM provider."
-  :group 'tools
+  :group 'ellama
   :type '(sexp :validate 'cl-struct-p))
 
+(defcustom ellama-providers nil
+  "LLM provider list for fast switching."
+  :group 'ellama
+  :type '(alist :key-type string))
+
 (defcustom ellama-spinner-type 'progress-bar
   "Spinner type for ellama."
-  :group 'tools
+  :group 'ellama
   :type `(choice ,@(mapcar
 		    (lambda (type)
 		      `(const ,(car type)))
 		    spinner-types)))
 
+(defun ellama-setup-keymap ()
+  "Set up the Ellama keymap and bindings."
+  (interactive)
+  (when (boundp 'ellama-keymap-prefix)
+    (defvar ellama-keymap (make-sparse-keymap)
+      "Keymap for Ellama Commands")
+
+    (define-key global-map (kbd ellama-keymap-prefix) ellama-keymap)
+
+    (let ((key-commands
+	   '(;; code
+	     ("c c" ellama-code-complete "Code complete")
+	     ("c a" ellama-code-add "Code add")
+	     ("c e" ellama-code-edit "Code edit")
+	     ("c i" ellama-code-improve "Code improve")
+	     ("c r" ellama-code-review "Code review")
+	     ;; summarize
+	     ("s s" ellama-summarize "Summarize")
+	     ("s w" ellama-summarize-webpage "Summarize webpage")
+	     ;; improve
+	     ("i w" ellama-improve-wording "Improve wording")
+	     ("i g" ellama-improve-grammar "Improve grammar and spelling")
+	     ("i c" ellama-improve-conciseness "Improve conciseness")
+	     ;; make
+	     ("m l" ellama-make-list "Make list")
+	     ("m t" ellama-make-table "Make table")
+	     ("m f" ellama-make-format "Make format")
+	     ;; ask
+	     ("a a" ellama-ask-about "Ask about")
+	     ("a i" ellama-ask-interactive "Ask interactively")
+	     ("a l" ellama-ask-line "Ask current line")
+	     ("a s" ellama-ask-selection "Ask selection")
+	     ;; text
+	     ("t t" ellama-translate "Text translate")
+	     ("t c" ellama-complete "Text complete")
+	     ;; define
+	     ("d w" ellama-define-word "Define word")
+	     ;; provider
+	     ("p s" ellama-provider-select "Provider select"))))
+      (dolist (key-command key-commands)
+	(define-key ellama-keymap (kbd (car key-command)) (cadr key-command))))))
+
 (defcustom ellama-keymap-prefix "C-c e"
   "Key sequence for Ellama Commands."
   :type 'string
-  :group 'tools)
+  :set (lambda (symbol value)
+	 (custom-set-default symbol value)
+	 (when value
+	   (ellama-setup-keymap)))
+  :group 'ellama)
+
+(defcustom ellama-ollama-binary (executable-find "ollama")
+  "Path to ollama binary."
+  :type 'string
+  :group 'ellama)
+
+(defcustom ellama-auto-scroll nil
+  "If enabled ellama buffer will scroll automatically during generation."
+  :type 'boolean
+  :group 'ellama)
+
+(defcustom ellama-fill-paragraphs '(text-mode)
+  "When to wrap paragraphs."
+  :group 'ellama
+  :type `(choice
+          (const :tag "Never fill paragraphs" nil)
+          (const :tag "Always fill paragraphs" t)
+          (function :tag "By predicate")
+          (repeat :tag "In specific modes" (symbol))))
 
 (defvar-local ellama--chat-prompt nil)
 
@@ -102,56 +176,21 @@
   (rx (minimal-match
        (literal "```") (zero-or-more anything))))
 
-(defun ellama-setup-keymap ()
-  "Set up the Ellama keymap and bindings."
-  (interactive)
-  (defvar ellama-keymap (make-sparse-keymap)
-    "Keymap for Ellama Commands")
-
-  (define-key global-map (kbd ellama-keymap-prefix) ellama-keymap)
-
-  (let ((key-commands
-         '(
-           ;; code
-	   ("c c" ellama-code-complete "Code complete")
-	   ("c a" ellama-code-add "Code add")
-	   ("c e" ellama-code-edit "Code edit")
-	   ("c i" ellama-code-improve "Code improve")
-	   ("c r" ellama-code-review "Code review")
-	   ;; summarize
-	   ("s s" ellama-summarize "Summarize")
-	   ("s w" ellama-summarize-webpage "Summarize webpage")
-	   ;; improve
-	   ("i w" ellama-improve-wording "Improve wording")
-	   ("i g" ellama-improve-grammar "Improve grammar and spelling")
-	   ("i c" ellama-improve-conciseness "Improve conciseness")
-	   ;; make
-	   ("m l" ellama-make-list "Make list")
-	   ("m t" ellama-make-table "Make table")
-	   ("m f" ellama-make-format "Make format")
-	   ;; ask
-	   ("a a" ellama-ask-about "Ask about")
-	   ("a i" ellama-ask-interactive "Ask interactively")
-	   ("a l" ellama-ask-line "Ask current line")
-	   ("a s" ellama-ask-selection "Ask selection")
-	   ;; text
-	   ("t t" ellama-translate "Text translate")
-	   ("t c" ellama-complete "Text complete")
-	   ;; define
-	   ("d w" ellama-define-word "Define word"))))
-    (dolist (key-command key-commands)
-      (define-key ellama-keymap (kbd (car key-command)) (cadr key-command)))))
+(defun ellama--code-filter (text)
+  "Filter code prefix/suffix from TEXT."
+  ;; Trim left first as `string-trim' trims from the right and ends up deleting all the code.
+  (string-trim-right (string-trim-left text ellama--code-prefix) ellama--code-suffix))
 
 (defcustom ellama-enable-keymap t
   "Enable or disable Ellama keymap."
   :type 'boolean
   :group 'ellama
   :set (lambda (symbol value)
-         (set symbol value)
-         (if value
-             (ellama-setup-keymap)
-           ;; If ellama-enable-keymap is nil, remove the key bindings
-           (define-key global-map (kbd ellama-keymap-prefix) nil))))
+	 (custom-set-default symbol value)
+	 (if value
+	     (ellama-setup-keymap)
+	   ;; If ellama-enable-keymap is nil, remove the key bindings
+	   (define-key global-map (kbd ellama-keymap-prefix) nil))))
 
 (defun ellama-stream (prompt &rest args)
   "Query ellama for PROMPT.
@@ -160,81 +199,76 @@ ARGS contains keys for fine control.
 :buffer BUFFER -- BUFFER is the buffer (or `buffer-name') to insert ellama reply
 in.  Default value is (current-buffer).
 
-:point POINT -- POINT is the point in buffer to insert ellama reaply at."
+:point POINT -- POINT is the point in buffer to insert ellama reply at.
+
+:filter FILTER -- FILTER is a function that's applied to (partial) response
+strings before they're inserted into the BUFFER.
+
+:session SESSION -- if session is non-nil, the PROMPT will be appended to the
+current (BUFFER-local) conversation.
+
+:on-error ON-ERROR -- ON-ERROR a function that's called with an error message on
+failure (with BUFFER current).
+
+:on-done ON-DONE -- ON-DONE a function that's called with the full response text
+when the request completes (with BUFFER current)."
   (let* ((buffer (or (plist-get args :buffer) (current-buffer)))
 	 (point (or (plist-get args :point)
-		    (with-current-buffer buffer (point)))))
+		    (with-current-buffer buffer (point))))
+	 (filter (or (plist-get args :filter) #'identity))
+	 (session (plist-get args :session))
+	 (errcb (or (plist-get args :on-error) (lambda (msg) (error "Error calling the LLM: %s" msg))))
+	 (donecb (or (plist-get args :on-done) #'ignore)))
     (with-current-buffer buffer
-      (unwind-protect
-	  (save-excursion
-	    (let* ((start (make-marker))
-		   (end (make-marker))
-		   (insert-text
-		    (lambda (text)
-		      ;; Erase and insert the new text between the marker cons.
-		      (with-current-buffer (marker-buffer start)
-			(save-excursion
-			  (goto-char start)
-			  (delete-region start end)
-			  (insert text)
-			  (fill-region start (point)))))))
-	      (setq ellama--change-group (prepare-change-group))
-	      (activate-change-group ellama--change-group)
-              (set-marker start point)
-              (set-marker end point)
-              (set-marker-insertion-type start nil)
-              (set-marker-insertion-type end t)
-	      (spinner-start ellama-spinner-type)
-	      (llm-chat-streaming ellama-provider
-				  (llm-make-simple-chat-prompt prompt)
-				  insert-text
-				  (lambda (text)
-				    (funcall insert-text text)
-				    (with-current-buffer buffer
-				      (undo-amalgamate-change-group ellama--change-group)
-				      (accept-change-group ellama--change-group)
-				      (spinner-stop)))
-				  (lambda (_ msg)
-				    (error "Error calling the LLM: %s" msg)
-				    (cancel-change-group ellama--change-group)))))))))
-
-(defun ellama-stream-filter (prompt prefix suffix buffer point)
-  "Query ellama for PROMPT with filtering.
-In BUFFER at POINT will be inserted result between PREFIX and SUFFIX."
-  (with-current-buffer buffer
-    (unwind-protect
-     (save-excursion
+      (if (and session ellama--chat-prompt)
+	  (llm-chat-prompt-append-response
+	   ellama--chat-prompt prompt)
+	(setq ellama--chat-prompt (llm-make-simple-chat-prompt prompt)))
       (let* ((start (make-marker))
-             (end (make-marker))
-	     (insert-text (lambda (text)
-			    ;; Erase and insert the new text between the marker cons.
-			    (with-current-buffer (marker-buffer start)
-			      (save-excursion
-				(goto-char start)
-				(delete-region start end)
-				;; remove prefix and suffix parts
-				(insert (string-trim-right
-					 (string-trim-left text prefix)
-					 suffix)))))))
+	     (end (make-marker))
+	     (insert-text
+	      (lambda (text)
+		;; Erase and insert the new text between the marker cons.
+		(with-current-buffer buffer
+		  ;; Manually save/restore point as save-excursion doesn't restore the point into
+		  ;; the middle of replaced text.
+		  (let ((pt (point)))
+		    (goto-char start)
+		    (delete-region start end)
+		    (insert (funcall filter text))
+                    (when (pcase ellama-fill-paragraphs
+                            ((cl-type function) (funcall ellama-fill-paragraphs))
+                            ((cl-type boolean) ellama-fill-paragraphs)
+                            ((cl-type list) (apply #'derived-mode-p ellama-fill-paragraphs)))
+                      (fill-region start (point)))
+		    (goto-char pt))
+		  (when-let ((ellama-auto-scroll)
+			     (window (get-buffer-window buffer)))
+		    (with-selected-window window
+		      (goto-char (point-max))
+		      (recenter -1)))))))
 	(setq ellama--change-group (prepare-change-group))
 	(activate-change-group ellama--change-group)
-        (set-marker start point)
-        (set-marker end point)
-        (set-marker-insertion-type start nil)
-        (set-marker-insertion-type end t)
+	(set-marker start point)
+	(set-marker end point)
+	(set-marker-insertion-type start nil)
+	(set-marker-insertion-type end t)
 	(spinner-start ellama-spinner-type)
 	(llm-chat-streaming ellama-provider
-			    (llm-make-simple-chat-prompt prompt)
+			    ellama--chat-prompt
 			    insert-text
 			    (lambda (text)
 			      (funcall insert-text text)
 			      (with-current-buffer buffer
 				(undo-amalgamate-change-group ellama--change-group)
 				(accept-change-group ellama--change-group)
-				(spinner-stop)))
+				(spinner-stop)
+				(funcall donecb text)))
 			    (lambda (_ msg)
-			      (cancel-change-group ellama--change-group)
-			      (error "Error calling the LLM: %s" msg))))))))
+			      (with-current-buffer buffer
+				(cancel-change-group ellama--change-group)
+				(spinner-stop)
+				(funcall errcb msg))))))))
 
 ;;;###autoload
 (defun ellama-chat (prompt)
@@ -244,50 +278,17 @@ In BUFFER at POINT will be inserted result between PREFIX and SUFFIX."
     (get-buffer-create ellama-buffer)
     (with-current-buffer ellama-buffer
       (funcall ellama-buffer-mode)))
+  (display-buffer ellama-buffer)
   (with-current-buffer ellama-buffer
-    (display-buffer ellama-buffer)
-    (if ellama--chat-prompt
-	(llm-chat-prompt-append-response
-	 ellama--chat-prompt prompt)
-      (setq ellama--chat-prompt (llm-make-simple-chat-prompt prompt)))
     (save-excursion
       (goto-char (point-max))
       (insert "## " ellama-user-nick ":\n" prompt "\n\n"
 	      "## " ellama-assistant-nick ":\n")
-      (let* ((start (make-marker))
-	     (end (make-marker))
-	     (point (point-max))
-	     (insert-text
-	      (lambda (text)
-		;; Erase and insert the new text between the marker cons.
-		(with-current-buffer (marker-buffer start)
-		  (save-excursion
-		    (goto-char start)
-		    (delete-region start end)
-		    (insert text)
-		    (fill-region start (point)))))))
-	(setq ellama--change-group (prepare-change-group))
-	(activate-change-group ellama--change-group)
-        (set-marker start point)
-        (set-marker end point)
-        (set-marker-insertion-type start nil)
-        (set-marker-insertion-type end t)
-	(spinner-start ellama-spinner-type)
-	(llm-chat-streaming ellama-provider
-			    ellama--chat-prompt
-			    insert-text
-			    (lambda (text)
-			      (funcall insert-text text)
-			      (with-current-buffer ellama-buffer
-				(save-excursion
-				  (goto-char (point-max))
-				  (insert "\n\n"))
-				(undo-amalgamate-change-group ellama--change-group)
-				(accept-change-group ellama--change-group)
-				(spinner-stop)))
-			    (lambda (_ msg)
-			      (cancel-change-group ellama--change-group)
-			      (error "Error calling the LLM: %s" msg)))))))
+      (ellama-stream prompt
+		     :session t
+		     :on-done (lambda (_) (save-excursion
+				      (goto-char (point-max))
+				      (insert "\n\n")))))))
 
 ;;;###autoload
 (defalias 'ellama-ask 'ellama-chat)
@@ -448,14 +449,12 @@ In BUFFER at POINT will be inserted result between PREFIX and SUFFIX."
 		(point-max)))
 	 (text (buffer-substring-no-properties beg end)))
     (kill-region beg end)
-    (ellama-stream-filter
+    (ellama-stream
      (format
       "Regarding the following code, %s, only ouput the result code in format ```language\n...\n```:\n```\n%s\n```"
       change text)
-     ellama--code-prefix
-     ellama--code-suffix
-     (current-buffer)
-     beg)))
+     :filter #'ellama--code-filter
+     :point beg)))
 
 ;;;###autoload
 (defun ellama-enhance-code ()
@@ -469,14 +468,12 @@ In BUFFER at POINT will be inserted result between PREFIX and SUFFIX."
 		(point-max)))
 	 (text (buffer-substring-no-properties beg end)))
     (kill-region beg end)
-    (ellama-stream-filter
+    (ellama-stream
      (format
       "Enhance the following code, only ouput the result code in format ```language\n...\n```:\n```\n%s\n```"
       text)
-     ellama--code-prefix
-     ellama--code-suffix
-     (current-buffer)
-     beg)))
+     :filter #'ellama--code-filter
+     :point beg)))
 
 ;;;###autoload
 (defun ellama-complete-code ()
@@ -487,16 +484,14 @@ In BUFFER at POINT will be inserted result between PREFIX and SUFFIX."
 		(point-min)))
 	 (end (if (region-active-p)
 		  (region-end)
-		(point-max)))
+		(point)))
 	 (text (buffer-substring-no-properties beg end)))
-    (ellama-stream-filter
+    (ellama-stream
      (format
       "Continue the following code, only write new code in format ```language\n...\n```:\n```\n%s\n```"
       text)
-     ellama--code-prefix
-     ellama--code-suffix
-     (current-buffer)
-     end)))
+     :filter #'ellama--code-filter
+     :point end)))
 
 ;;;###autoload
 (defun ellama-add-code (description)
@@ -511,20 +506,17 @@ buffer."
 		  (region-end)
 		(point-max)))
 	 (text (buffer-substring-no-properties beg end)))
-    (ellama-stream-filter
+    (ellama-stream
      (format
       "Context: \n```\n%s\n```\nBased on this context, %s, only ouput the result in format ```\n...\n```"
       text description)
-     ellama--code-prefix
-     ellama--code-suffix
-     (current-buffer)
-     end)))
+     :filter #'ellama--code-filter)))
 
 
 ;;;###autoload
 (defun ellama-render (needed-format)
   "Render selected text or text in current buffer as NEEDED-FORMAT."
-  (interactive)
+  (interactive "sSpecify required format: ")
   (let* ((beg (if (region-active-p)
 		  (region-beginning)
 		(point-min)))
@@ -568,5 +560,32 @@ buffer."
       (kill-region (point) (point-max))
       (ellama-summarize))))
 
+(defun ellama-get-ollama-local-model ()
+  "Return llm provider for interactively selected ollama model."
+  (interactive)
+  (let ((model-name
+	 (completing-read "Select ollama model: "
+			  (mapcar (lambda (s)
+				    (car (split-string s)))
+				  (seq-drop
+				   (process-lines ellama-ollama-binary "ls") 1)))))
+    (make-llm-ollama
+     :chat-model model-name :embedding-model model-name)))
+
+;;;###autoload
+(defun ellama-provider-select ()
+  "Select ellama provider."
+  (interactive)
+  (let* ((providers (if (and ellama-ollama-binary
+			     (file-exists-p ellama-ollama-binary))
+			(push '("ollama model" . (ellama-get-ollama-local-model))
+			      ellama-providers)
+		      ellama-providers))
+	 (variants (mapcar #'car providers)))
+    (setq ellama-provider
+	  (eval (alist-get
+		 (completing-read "Select model: " variants)
+		 providers nil nil #'string=)))))
+
 (provide 'ellama)
 ;;; ellama.el ends here.
